#include "arm.h"
#include "entry.h"
#include "safeg_syscalls.h"
#include "safeg_measures.h"

/*
 * -----------------
 * Workspace Offsets
 * -----------------
 * Offsets within the workspace
 */

/*
 * Offset to first banked registers from sp_mon.
 * Calculated using the N World Workspace but values are the same for S World.
 */
#define BankedOffset    (N_IRQ_SPSR - NtContextBase)

/*
 * --------------------
 * Workspace Constants
 * --------------------
 */

#define WORKSPACE_RESET_VALUE   0x0

/*
 * Interrupt Related Constants
 */
#define HIGHVEC_IRQADR        0xffff0018

/*
 * NORMAL_CPSR_INIT_VALUE
 */
#define NORMAL_CPSR_INIT_VALUE CPSR_BIT_8_A | CPSR_BIT_7_I | CPSR_MODE_SVC

/*
 * NORMAL CP15 CTRL 쥸ΥꥻåȻ
 */
#define NORMAL_CP15_CTRL_INIT_VALUE 0x50078

/*
 * ----------------
 * MACRO: SaveState
 * ----------------
 * Saves registers to the workspace sp points into.
 * Corrupts r6, r7, r12
 * Trust(SCRNSӥåȤ0)ǸƤӤ.
 */
    .macro SaveState
        stmdb   sp, {r0-lr}^
        add     r6, sp, #BankedOffset
        cps     #CPSR_MODE_IRQ
        mrs     r12, spsr
        stmia   r6!, {r12-lr}
        cps     #CPSR_MODE_SVC
        mrs     r12, spsr
        stmia   r6!, {r12-lr}
        cps     #CPSR_MODE_ABORT
        mrs     r12, spsr
        stmia   r6!, {r12-lr}
        cps     #CPSR_MODE_UNDEF
        mrs     r12, spsr
        stmia   r6!, {r12-lr}
        cps     #CPSR_MODE_FIQ
        mrs     r7, spsr
        stmia   r6, {r7-lr}
        cps     #CPSR_MODE_MONITOR
    .endm

/*
 * ----------------
 * MACRO: LoadState
 * ----------------
 * SPϸƤӽФƥȤξؤƤ뤳.
 * Trust(SCRNSӥåȤ0)ǸƤӤ.
 * λϥ˥⡼ɤˤʤäƤ, 쥸ˤϳƥƥȤƤƤ뤿,
 * LR_mon, SP_monΤ߻Ѳǽ.
 */

    .macro LoadState
        add     r6, sp, #BankedOffset
        cps     #CPSR_MODE_IRQ
        ldmia   r6!, {r12-lr}
        msr     spsr_fsxc, r12
        cps     #CPSR_MODE_SVC
        ldmia   r6!, {r12-lr}
        msr     spsr_fsxc, r12
        cps     #CPSR_MODE_ABORT
        ldmia   r6!, {r12-lr}
        msr     spsr_fsxc, r12
        cps     #CPSR_MODE_UNDEF
        ldmia   r6!, {r12-lr}
        msr     spsr_fsxc, r12
        cps     #CPSR_MODE_FIQ
        ldmia   r6, {r7-lr}
        msr     spsr_fsxc, r7
        cps     #CPSR_MODE_MONITOR
        ldmdb   sp, {r0-lr}^
    .endm

/*
 * ----------------------------------------------------------------------------
 * ===============
 * ˥٥
 * ===============
 *
 * table, ٥0Ϥ֤
 */
    .section .vector, "a"
    .align 5
    .globl MonitorVectorTableBase
MonitorVectorTableBase:
    ldr pc, reset_vector       /* ꥻå             */
    ldr pc, undef_vector       /* ̤̿           */
    ldr pc, smc_vector         /* SMC                  */
    ldr pc, prefech_vector     /* ץեåܡ */
    ldr pc, data_abort_vector  /* ǡܡ       */
    ldr pc, reset_vector       /* ͽ                 */
    ldr pc, irq_vector         /* IRQ                  */
    ldr pc, fiq_vector         /* FIQ                  */

/*
 *  㳰٥̿ᤫ黲Ȥ
 *  襢ɥ쥹
 */
    .global monitor_vector_ref_tbl
monitor_vector_ref_tbl:
reset_vector:
    .long   disaster
undef_vector:
    .long   disaster
smc_vector:
    .long   smc_handler
prefech_vector:
    .long   disaster
data_abort_vector:
    .long   data_abort_handler
irq_vector:
    .long   irq_handler
fiq_vector:
    .long   fiq_handler
disaster: //error handling 
    b   disaster

/*
 *-----------------------------------------------------------------------------
 * ===========
 * IRQ Handler
 *
 * IRQNon Trusted¦餷ʤȲꤷƤ
 * ˥⡼, IRQ/FIQػߤǸƤӽФ.
 * IRQ ν򤷤Ƥ֤FIQSMC뤳ȤΤ,
 * IRQ ѤΥå̤ˤ.
 *
 * H offset
 *    +0x14 _smm_stack
 *   ip        +0x10
 *   r3        +0x0c
 *   r2        +0x08
 *   r1        +0x04
 * spr0        +0x00
 * L 
 * ===========
 */
    .text
    .align 5
    .global irq_handler
irq_handler:
    /* ȯΥɤȽǤ, Trust褿Ȥб򤳤Ǥ */
    CP15_SCR_READ sp
    tst sp, #SCR_BIT_0_NS
    beq trusted_irq                 // trustξ, trusted_irqإ

    /* åڤؤ, ƥȤ¸.
     * fdǤΤޤѤǤץǥȤΤ־϶.
     */
    ldr sp, =_smm_stack
    ldr sp, [sp]
    stmfd sp!, {r0-r1}

#if SAFEG_MEASURE_IRQ_OVH == 1
#if SAFEG_MEASURE_GPIO_ENA == 1
    gpio_set r0, r1, GPIO_0_DATA_MASK_0  /* set port 0 */
#endif /* SAFEG_MEASURE_GPIO_ENA */
#if SAFEG_MEASURE_PERF_ENA == 1
    perfmon_start r0
#endif /* SAFEG_MEASURE_PERF_ENA */
#endif /* SAFEG_MEASURE_IRQ_OVH */

#if LINUX_MEASURE_TIMER_LAT == 1
#if SAFEG_MEASURE_PERF_ENA == 1
    perfmon_start r0
#endif /* SAFEG_MEASURE_PERF_ENA */
#endif /* SAFEG_MEASURE_LINUX_TIMER_LAT */

    /*
     * NT¦Υ٥ơ֥ɤ߽Ф.
     * SCRNSӥåȤ1ξ֤ɤ߽Фɬפ.
     * (NT¦MMUѤ).
     */
    CP15_CONTROL_READ r0
    tst r0, #CP15_CONTROL_V_BIT

    /* high vector ξ r0  HIGHVEC_IRQ ɤ߹ */ 
    ldrne r0, =HIGHVEC_IRQADR
    bne highvec

    /* high vector Ǥʤ */
    CP15_VBAR_READ r0                // check non-secure irq vector address
    add r0, r0, #0x18                // IRQ٥Υɥ쥹 vectorbase + 0x18
highvec:
    /*
     * ޤǤNSǼ¹Ԥɬפ.
     * SCRNSӥåȤ0ˤƤ.
     * (⡼ؤԤ). 
     */
    mov r1, #SCR_S
    CP15_SCR_WRITE r1

    /*
     * Υϥɥ餬ä˥פºݤΥ٥ɥ쥹¸.
     * ¸쥸ѤȥեåȤѤΤդ뤳.
     *
     * Note: the stack after "stmfd sp!, {r0-r1}" is like this:
     *
     *                                 +------------------+ high address
     * _smm_stack = old_sp = sp + 8 -> |      xxx         |
     *                       sp + 4 -> |      r1          |     
     *                       sp + 0 -> |      r0          |
     *                                 +------------------+ low address
     *
     * The xxx space is there because we used "!" which increments the
     * stack pointer before. Now we will do: xxx = r0 = irq handler address
     */
    str r0, [sp, #8]

    /*
     * NT¦γߥϥɥѤ˺Ѥ褦,
     * lr_mon, spsr_mon  lr_irq, spsr_irq ˰ܤ
     */
    mov r0, lr                       //[r0] = lr_mon
    mrs r1, spsr                     //[r1] = spsr_mon
    cps #CPSR_MODE_IRQ
    mov lr, r0                       // lr_irq = lr_mon
    msr spsr_cxsf, r1                // spsr_irq = spsr_mon
    cps #CPSR_MODE_MONITOR

    /* NT ¦ IRQ ٥ƤӽФν */
    mov r1, #SCR_NS
    CP15_SCR_WRITE r1

    /* spsr  IRQ ⡼ɤȤʤ褦򤹤 */
    mov r0, #(CPSR_MODE_IRQ | CPSR_BIT_7_I)
    msr spsr_fsxc, r0

#ifdef ENABLE_IRQ_MANAGEMENT
    stmfd sp!, {r2-r3, ip}
    bl irq_management
    ldmfd sp!, {r2-r3, ip}
#endif

#if SAFEG_MEASURE_IRQ_OVH == 1
#if SAFEG_MEASURE_GPIO_ENA == 1
    gpio_clear r0, r1, GPIO_0_DATA_MASK_0  /* clear port 0 */
#endif /* SAFEG_MEASURE_GPIO_ENA */
#if SAFEG_MEASURE_PERF_ENA == 1
    stmfd sp!, {r2}
    perfmon_end r0, r1, r2
    ldmfd sp!, {r2}
#endif /* SAFEG_MEASURE_PERF_ENA */
#endif /* SAFEG_MEASURE_IRQ_OVH */

    /* NT ¦ IRQ ٥˥ */
    ldmfd sp!, {r0-r1, pc}^

   /*
    * trusted_irq ξνԤ.
    * ȼ˽Ԥˤ, trusted_irq_handler .
    */
trusted_irq:
    ldr r0, =trusted_irq_handler
    ldr r0, [r0]
    cmp r0, #0
    movne pc, r0
    b   disaster

/*
 *-----------------------------------------------------------------------------
 * ===========
 * FIQ handler
 * ===========
 */
    .text
    .align 5
    .global fiq_handler
fiq_handler:
    /* ȯΥɤȽ */
    /* Note: r13=sp and r14=lr are banked in monitor mode */
    CP15_SCR_READ sp
    tst sp, #SCR_BIT_0_NS
    bne non_trusted_fiq

    /* secure褿 */
trusted_fiq:
    /* save trust r0-r1 registers in SafeG stack cause we need to use them */
    ldr sp, =_smm_stack
    ldr sp, [sp]
    stmfd sp!, {r0-r1}

#if SAFEG_MEASURE_FIQ_FIQ == 1
#if SAFEG_MEASURE_GPIO_ENA == 1
    gpio_set r0, r1, GPIO_0_DATA_MASK_0  /* set port 0 */
#endif /* SAFEG_MEASURE_GPIO_ENA */
#if SAFEG_MEASURE_PERF_ENA == 1
    perfmon_start r0
#endif /* SAFEG_MEASURE_PERF_ENA */
#endif /* SAFEG_MEASURE_FIQ_FIQ */

    /*
     * T¦γߥϥɥѤ˺Ѥ褦,
     * lr_mon, spsr_mon  lr_irq, spsr_irq ˰ܤ
     */
    mov r0, lr
    mrs r1, spsr
    cps #CPSR_MODE_IRQ
    mov lr, r0
    msr spsr_fsxc, r1
    cps #CPSR_MODE_MONITOR

#ifdef ENABLE_FIQ_MANAGEMENT
    stmfd sp!, {r2-r3, ip}
    bl fiq_management
    ldmfd sp!, {r2-r3, ip}
#endif

    ldmfd sp!, {r0-r1}
    b fiq_exit

    /* non-secure褿 */
non_trusted_fiq:
    /* SaveStateSCRsecureѹ */
    mov sp, #SCR_S
    CP15_SCR_WRITE sp
    /* ߤäƤΤ, 襢ɥ쥹LR4餷ͤˤʤ */
    sub lr, lr, #4
    /* save non-trust pc and spsr to monitor stack */
    ldr sp, =NtContextBase
    srsia #CPSR_MODE_MONITOR
    /* save the rest of non-trust registers */
    SaveState
    /*
     * LoadState 򤹤IRQ⡼ɤƤؤΤǽ񤭴˹Ԥ.
     * LoadState 륳ƥȤ,  SMC ¹Ի¸줿
     * Ǥ뤿, IRQϥɥǻѤ뤿˻ 4 ­Ƥ.
     * Trust¦IRQϥɥ餬ä֤뤿 SPSR_irq, LR_irq 
     * IRQ⡼ɤꤷƤ.
     */
    ldr sp, =TContextBase
    ldmfd sp, {r0, r1} // SMCƤӽФ pc, cpsr.
    add r0, r0, #4
    str r1, [sp, #0xc]
    str r0, [sp, #0x14]
    LoadState
    /* Prepare stack to store CPSR and PC before returning with rfeia*/
    ldr sp, =_smm_stack
    ldr sp, [sp]

#ifdef ENABLE_FIQ_MANAGEMENT
    stmfd sp!, {r0-r3, ip}
    bl fiq_management
    ldmfd sp!, {r0-r3, ip}
#endif

/*
 * secureɬȴΤǡˤSCRsecureǤ.
 * rfeia뤿˥åCPSRPCѤ.
 */
fiq_exit:
    /* push CPSR = (IRQ and FIQ disabled) + IRQ mode */
    mov lr, #(CPSR_MODE_IRQ | CPSR_BIT_6_F | CPSR_BIT_7_I)
    str lr, [sp, #-4]!
    /* IRQ٥Υɥ쥹vectorbase + 0x18 */
    CP15_VBAR_READ lr
    add lr, lr, #0x18
    str lr, [sp, #-4]!
    /* IRQ⡼ɤڤؤ (load CPSR + PC from stack) T¦ IRQ٥˥ */
    rfeia sp

/*
 * ===========
 * SMC handler
 * ===========
 */
    .text
    .align 5
    .global smc_handler
smc_handler:
    /* check if it is a Trust or Non-Trust SMC (NS bit) */
    CP15_SCR_READ sp
    tst sp, #SCR_BIT_0_NS

    /* if is Non-Trust SMC jump to nt_smc_handler */
    bne nt_smc_handler

t_smc_handler:
    /* otherwise its a Trust SMC so first we save the Trust context */
    ldr sp, =TContextBase
    srsia   #CPSR_MODE_MONITOR  // [sp_mon] = lr_mon, [sp_mon, #4] = spsr_mon
    SaveState                   // S-Context = Context

    /* r0 has the syscall, branch to the appropriate SMC handler */
    cmp r0, #T_SMC_REBOOT
    beq t_smc_reboot

t_smc_switch:
    /* load the non-trust context state */
    ldr sp, =NtContextBase
    LoadState                   // Context = N-Context

    /* set the SCR to non-trust world settings */
    mov lr, #SCR_NS             // lr_mon = Normal SCR Config
    CP15_SCR_WRITE lr           // cp15_SCR = lr_mon (N-World Config)

    /* flush the branch target address cache */
    CP15_Flush_BTAC lr          // flush BTAC (corrupts lr_mon)

t_smc_switch_exit:
    /* switch to non-trust world */
    rfeia   sp                  // load cpsr and pc from NtContext

t_smc_reboot: /* TODO: Test this! */
    /*
     * r1 ˤϥΥɥ쥹äƤ
     * reset  CPSR  PC NT¦ƥȤꤹ
     * r0  NORMAL_CP15_CTRL_INIT_VALUE Υɤ˻ȤʤΤ,
     * stmea ȥפʤ.
     */
    ldr r4, =NtContextBase
    ldr r2, =NORMAL_CPSR_INIT_VALUE
    stmea r4, {r1, r2}

    /* ľ linux  boot 뤿Υ쥸 */
    mov r1, #0
    mov r2, #0x5e0
    mov r3, #0x100
    ldr r4, =NtContextSavedArea
    stmea r4, {r1-r3}

    /* CP15 CTRL 쥸 Non-Secure ˤƤɬפ. */
    mov     r2, #SCR_NS
    CP15_SCR_WRITE  r2

    /* MMU  åDISABLEˤ */
    ldr r1, =NORMAL_CP15_CTRL_INIT_VALUE
    mcr p15, 0, r1, c1, c0, 0

    /* ä Secure᤹. */
    mov     r2, #SCR_S
    CP15_SCR_WRITE  r2
    b t_smc_switch

nt_smc_handler:
    /* switch to NT (and pass arguments) */
    cmp r0, #NT_SMC_SWITCH
    beq nt_smc_switch
    /* emulate a software interrupt in Trust world */
    cmp r0, #NT_SMC_SOFTIRQ
    addeq lr, lr, #4
    beq non_trusted_fiq
    /* invalidate the cache */
    cmp r0, #NT_SMC_INV_CACHE
    beq nt_smc_cache

    /* if SMC number is not supported go back to nt with error (TODO: test) */
    mov r1, #NT_SMC_SWITCH_ERR
    b nt_smc_return_to_nt

nt_smc_switch:
    /* go to secure world, to save NT state in secure memory */
    mov     sp, #SCR_S          // lr_mon = Secure World SCR config
    CP15_SCR_WRITE  sp          // cp15_SCR = lr_mon (S-World Config)

    /* save NT world state */
    ldr sp, =NtContextBase
    srsia #CPSR_MODE_MONITOR
    SaveState                   // N-Context = Context

    cmp r1, #T_SMC_SWITCH_HANDLER
    beq nt_smc_switch_raise_handler

    /*
     * In S_USR_R1 we have  &ret_args, get it to r6 (SaveState already
     * corrupted r6,r7,r12) and write args (r1-r2) in there
     */
    ldr r6, =TContextSavedArea  // r6 = TContextSavedArea
    ldr r6, [r6, #4]            // r6 = *(S_USR_R1) = ret_args
    str r1, [r6, #0]            // ret_args->arg0 = r1
    str r2, [r6, #4]            // ret_args->arg1 = r2

nt_smc_switch_exit:
    /* Load the trust world state */
    /* Note: BTAC flush is automatic in this direction */
    ldr     sp, =TContextBase
    LoadState                   // Context = S-Context
    rfeia   sp                  // switch to the trust-world os

nt_smc_switch_raise_handler:
    /* turn to trust world and raise a safeg soft interrupt handler */
    ldr sp, =TContextBase
    ldmfd sp, {r0, r1} // SMCƤӽФ pc, cpsr.
    add r0, r0, #4
    str r1, [sp, #0xc]
    str r0, [sp, #0x14]
    LoadState
    ldr sp, =_smm_stack
    ldr sp, [sp]
    mov lr, #(CPSR_MODE_IRQ | CPSR_BIT_6_F | CPSR_BIT_7_I)
    str lr, [sp, #-4]!
    /* SAFEG٥Υɥ쥹vectorbase + 0x20 */
    CP15_VBAR_READ lr
    add lr, lr, #0x20
    str lr, [sp, #-4]!
    /* IRQ ⡼ɤڤؤ, T ¦ IRQ ٥˥ */
    rfeia sp

nt_smc_cache:
    /* check arg0 in r1 to know what to invalidate */
    cmp r1, #NT_SMC_INV_CACHE_D
    beq nt_smc_cache_invalidate_dcache
    cmp r1, #NT_SMC_INV_CACHE_I
    beq nt_smc_cache_invalidate_icache
    cmp r1, #NT_SMC_INV_CACHE_ALL
    beq nt_smc_cache_invalidate_allcache

    /* if argument number is not supported return with error (TODO: test)*/
    mov r1, #NT_SMC_INV_CACHE_ERR
    b nt_smc_return_to_nt

nt_smc_cache_invalidate_dcache:
    /* invalidate the entire data cache */
    mov r0, #0
    mcr p15, 0, r0, c7, c6, 0
    b nt_smc_return_to_nt

nt_smc_cache_invalidate_icache:
    /* invalidate the entire instruction cache */
    mov r0, #0
    mcr p15, 0, r0, c7, c5, 0
    b nt_smc_return_to_nt

nt_smc_cache_invalidate_allcache:
    /* invalidate both the data and the instruction cache */
    mov r0, #0
    mcr p15, 0, r0, c7, c14, 0
    mcr p15, 0, r0, c7, c5, 0
    b nt_smc_return_to_nt

nt_smc_return_to_nt:
    movs pc, lr

/*
 * SCREAӥåȤΩƤȤˤäơNT¦Trustΰ˥ݤ
 * Υ٥ƤФ롥
 * ʤΤǳμ¤NT¦Ȥdata_abort_handlerϸƤӽФ롥
 * NT¦ˤʤΤǡSaveStateϸƤӽФʤ
 * ƤȤƤdata abortNT¦ΥեȤϥɥ󥰤Ȥ
 * 塤SWIȤT¦˽뤳ȤȤ
 */
    .text
    .align 5
    .global data_abort_handler
data_abort_handler:
    /* LoadStateƤӽФ, SecureˤƤɬפ. */
    mov     sp, #SCR_S                 // lr_mon = Secure World SCR config
    CP15_SCR_WRITE  sp                 // cp15_SCR = lr_mon (S-World Config)

    /*
     * TContextBase, TContextBase+4   SMC ƤӽФ줿
     * Ȥ pc, cpsr ȤʤäƤ롥
     * ͤ swi λȤʤΤǡswi ̾ƤФ줿֤˹碌
     * lr_svc, spsr_svc ꤹ롥
     * lr_svc, spsr_svc˲뤬asm_target_idleǤϤ˲Ƥ
     * 褤ȤˤʤäƤΤ¸Ϥʤ
     */
    ldr sp, =TContextBase
    ldmia sp, {r0, r1}
    ldr sp, =S_SVC_LR
    str r0, [sp]
    str r1, [sp, #-8]

    /*
     * rfeia뤿 S_PC, S_CPSR 񤭴롥
     * FIQ ġIRQ ػߡSVC⡼ɤˤʤ֤ȴ. 
     */
    ldr sp, =TContextBase
    /* SWI ٥Υɥ쥹vectorbase + 0x8 */
    CP15_VBAR_READ r0
    add r0, r0, #0x8
    mov r1, #(CPSR_MODE_SVC | CPSR_BIT_7_I)
    stmia sp, {r0, r1}

    /*
     * TƥȤ
     */
    LoadState

    /* SVC ⡼ɤڤؤ, T ¦ SWI ٥˥ */
    rfeia sp

/*
 * SafeG εưԤ.
 * trust¦SMCˤnon-trust˰ưǤ褦
 * non-trustεưΤξNTƥȤ¸.
 * Trust¦εưƤӽФ.
 *
 * r0: Non-Trust¦ȥɥ쥹
 * r1: Trust¦ȥɥ쥹
 */
    .text
    .align 5
    .global SafeGStart
SafeGStart:
    /* Non-Trust¦ΥƥȤꤹ */
    // stmea  label Ȥäñȯȥ2Ĥˤ
    ldr r2, =NtContextBase
    ldr r3, =NORMAL_CPSR_INIT_VALUE
    stmea r2, {r0, r3}

    /* Trust¦ꥻåȻΥ⡼ɤѹ */
    cps #CPSR_MODE_SVC
    
    /* Trust¦Υȥɥ쥹˥פ */
    mov pc, r1


/*
 *-----------------------------------------------------------------------------
 * =================
 * Monitor workspace
 * =================
 */
    .bss
    .align 5
    .global NtContextSavedArea
    .global NtContextSavedAreaEnd
    .global TContextSavedArea
    .global TContextSavedAreaEnd
    .global NtContextBase
    .global TContextBase
NtContextSavedArea:
N_USR_R0       : .long 0x00
N_USR_R1       : .long 0x00
N_USR_R2       : .long 0x00
N_USR_R3       : .long 0x00
N_USR_R4       : .long 0x00
N_USR_R5       : .long 0x00
N_USR_R6       : .long 0x00
N_USR_R7       : .long 0x00
N_USR_R8       : .long 0x00
N_USR_R9       : .long 0x00
N_USR_R10      : .long 0x00
N_USR_R11      : .long 0x00
N_USR_R12      : .long 0x00
N_USR_SP       : .long 0x00
N_USR_LR       : .long 0x00

NtContextBase:
N_PC           : .long 0x00    // PC to load to return to Normal World
N_CPSR         : .long 0x00    // CPSR to load when returning to Normal World
N_SCRATCH      : .long 0x00
N_IRQ_SPSR     : .long 0x00    // Normal World IRQ mode SPSR, SP and LR reg
N_IRQ_SP       : .long 0x00
N_IRQ_LR       : .long 0x00
N_SVC_SPSR     : .long 0x00    // etc.
N_SVC_SP       : .long 0x00
N_SVC_LR       : .long 0x00
N_ABT_SPSR     : .long 0x00
N_ABT_SP       : .long 0x00
N_ABT_LR       : .long 0x00
N_UND_SPSR     : .long 0x00
N_UND_SP       : .long 0x00
N_UND_LR       : .long 0x00
N_FIQ_SPSR     : .long 0x00
N_FIQ_R8       : .long 0x00
N_FIQ_R9       : .long 0x00
N_FIQ_R10      : .long 0x00
N_FIQ_R11      : .long 0x00
N_FIQ_R12      : .long 0x00
N_FIQ_SP       : .long 0x00
N_FIQ_LR       : .long 0x00
NtContextSavedAreaEnd:

TContextSavedArea:
S_USR_R0       : .long 0x00    // Secure World USR/SYS registers
S_USR_R1       : .long 0x00
S_USR_R2       : .long 0x00
S_USR_R3       : .long 0x00
S_USR_R4       : .long 0x00
S_USR_R5       : .long 0x00
S_USR_R6       : .long 0x00
S_USR_R7       : .long 0x00
S_USR_R8       : .long 0x00
S_USR_R9       : .long 0x00
S_USR_R10      : .long 0x00
S_USR_R11      : .long 0x00
S_USR_R12      : .long 0x00
S_USR_SP       : .long 0x00
S_USR_LR       : .long 0x00

TContextBase:
S_PC           : .long 0x00    // PC to load to return to Secure World
S_CPSR         : .long 0x00    // CPSR to load to return to Secure World
S_SCRATCH      : .long 0x00
S_IRQ_SPSR     : .long 0x00    // Secure World IRQ mode SPSR, SP and LR reg
S_IRQ_SP       : .long 0x00
S_IRQ_LR       : .long 0x00
S_SVC_SPSR     : .long 0x00    // etc.
S_SVC_SP       : .long 0x00
S_SVC_LR       : .long 0x00
S_ABT_SPSR     : .long 0x00
S_ABT_SP       : .long 0x00
S_ABT_LR       : .long 0x00
S_UND_SPSR     : .long 0x00
S_UND_SP       : .long 0x00
S_UND_LR       : .long 0x00
S_FIQ_SPSR     : .long 0x00
S_FIQ_R8       : .long 0x00
S_FIQ_R9       : .long 0x00
S_FIQ_R10      : .long 0x00
S_FIQ_R11      : .long 0x00
S_FIQ_R12      : .long 0x00
S_FIQ_SP       : .long 0x00
S_FIQ_LR       : .long 0x00
TContextSavedAreaEnd:
